"""Module containing global configuration options.

Attributes
----------
ARROW_STYLE : str
    definition of TikZ arrow style. Any style defined in the TikZ `arrows`
    library is valid, e.g. `'->'` and `'-latex'`. The default `''` gives a line without
    arrow head.
SERIF_FONT : bool
    `False` to use sans-serif font (default), `True` to use serif fonts
"""

ARROW_STYLE: str = ""

SERIF_FONT: bool = False



"""Tests for `Diagram` class."""

from collections.abc import Generator
from os import chdir

import pytest
from pytest import FixtureRequest

from pyrbd import Block, Diagram, Group, Series, config


@pytest.fixture(name="arrow_style", scope="module", params=["", "-latex"])
def arrow_style_fixture(request: FixtureRequest) -> Generator[str, None, None]:
    """Arrow style pytest fixture."""

    styles: str = request.param

    yield styles


@pytest.fixture(name="diagram")
def diagram_fixture(arrow_style: str) -> Diagram:
    """Diagram pytest fixture."""

    config.ARROW_STYLE = arrow_style

    start_block = Block("Start", "myblue", parent=None)
    parallel = 2 * Block("Parallel blocks", "gray", parent=start_block)
    group = Group(
        [
            Block(r"Block 1", "yellow!50") + Block(r"Block 2", "yellow!50"),
            Block(r"Block 3", "yellow!50")
            + Block(r"Block 4", "yellow!50")
            + Block(r"Block 5", "yellow!50"),
        ],
        parent=parallel,
        text="Group",
        color="yellow",
    )
    series = Series(
        [Block(r"Block A", "orange!50"), Block(r"Block B", "orange!50")],
        "Series",
        "orange",
        parent=group,
    )
    end_block = Block("End", "green!50", parent=series)

    return Diagram(
        "test_diagram",
        [start_block, parallel, group, series, end_block],
        "Fire",
        colors={"myblue": "8888ff"},
    )


def test_diagram_init(diagram: Diagram) -> None:
    """Test __init__ of `Diagram` class."""

    assert diagram.filename == "test_diagram"
    assert "myblue" in diagram.colors
    assert isinstance(diagram.head, Block)


def test_diagram_wo_hazard() -> None:
    """Test __init__ of `Diagram` class without `hazard` specified."""

    head = Block("block1", "white")
    blocks = [head, Block("block2", "blue"), Block("block3", "green")]
    diagram = Diagram("test_diagram", blocks)

    # When hazard is not specified, blocks[0] is set as head of diagram
    assert diagram.head is head
    assert len(diagram.blocks) == 2


def test_diagram_write(tmp_path, diagram: Diagram) -> None:
    """Test `Diagram` `write` method."""

    temp_dir = tmp_path / "test_diagram"
    temp_dir.mkdir()

    chdir(temp_dir)

    diagram.write()

    tmp_file = temp_dir / f"{diagram.filename}.tex"

    for hex_code in diagram.colors.values():
        assert hex_code in tmp_file.read_text()
    assert config.ARROW_STYLE in tmp_file.read_text()
    assert diagram.head.text in tmp_file.read_text()



"""Module containing Diagram class definition."""

import subprocess

import pymupdf

from . import config
from .block import Block, Group
from .templates import JINJA_ENV


class Diagram:
    """Reliability block diagram class definition.

    Parameters
    ----------
    name : str
        name of diagram
    blocks : list[Block]
        list of `Block` instances
    hazard : str, default=""
        string defining the `hazard` block text (optional)
    colors : dict[str, str] | None, default=None
        dictionary with custom color definitions in HEX format:
        `{'color name': '6 digit hex code'}`

    Attributes
    ----------
    colors : dict[str, str], default={"arrowcolor": "4c4d4c", "hazardcolor": "ff6666"}
        default diagram color definitions
    """

    _template: str = "diagram.tex.jinja"
    colors: dict[str, str] = {"arrowcolor": "4c4d4c", "hazardcolor": "ff6666"}

    def __init__(
        self,
        name: str,
        blocks: list[Block],
        hazard: str = "",
        colors: dict[str, str] | None = None,
    ) -> None:
        self.filename = name
        if hazard:
            self.head = Block(hazard, "hazardcolor")
        else:
            self.head = blocks.pop(0)

        self.head.id = "0"
        self.blocks = blocks
        self.blocks[0].parent = self.head

        if colors is not None:
            self.colors = self.colors | colors

    def write(self) -> None:
        """Write diagram to .tex file."""

        environment = JINJA_ENV
        template = environment.get_template(self._template)

        context = {
            "serif_font": config.SERIF_FONT,
            "arrow_style": config.ARROW_STYLE,
            "color_defs": [{"name": name, "hex_code": code} for name, code in self.colors.items()],
            "blocks": list(block.get_node() for block in [self.head, *self.blocks]),
            "final_block": self.blocks[-1] if isinstance(self.blocks[-1], Group) else None,
        }
        content = template.render(context)

        with open(f"{self.filename}.tex", mode="w", encoding="utf-8") as file:
            file.write(content)

    def _to_svg(self) -> str:
        """Convert diagram file from pdf to svg.

        Returns
        -------
        str
            filename of .svg file
        """

        pdf_document = pymupdf.open(f"{self.filename}.pdf")
        page = pdf_document[0]

        # Get and convert page to svg image
        svg_content = page.get_svg_image().splitlines()
        svg_content.insert(
            1,
            "\n".join(
                [
                    "<style>",
                    "   @media (prefers-color-scheme: light) { :root { --color: #000000; } }",
                    "   @media (prefers-color-scheme: dark) { :root { --color: #DDDDDD; } }",
                    "</style>",
                ]
            ),
        )
        svg_content = "\n".join(svg_content).replace(r"#4c4d4c", "var(--color)")

        # Save to file
        with open(output_file := f"{self.filename}.svg", "w", encoding="utf-8") as file:
            file.write(svg_content)

        pdf_document.close()

        return output_file

    def _to_png(self) -> str:
        """Convert diagram file from pdf to png.

        Returns
        -------
        str
            filename of .png file
        """

        pdf_document = pymupdf.open(f"{self.filename}.pdf")
        page = pdf_document[0]

        # Get image
        image = page.get_pixmap(dpi=300)

        # Save to file
        image.save(output_file := f"{self.filename}.png")

        pdf_document.close()

        return output_file

    def compile(self, output: str | list[str] = "pdf", clear_source: bool = True) -> list[str]:
        """Compile diagram .tex file.

        Parameters
        ----------
        output : str | list[str], default='pdf'
            output format string or list of output formats for diagram. Valid output formats are

            - `'pdf'` (default)
            - `'svg'`
            - `'png'`

        clear_source : bool, default=True
            .tex source file is deleted after compilation if `True`

        Returns
        -------
        list[str]
            list of output filenames

        Raises
        ------
        FileNotFoundError
            If .tex file is not found, e.g. because `Diagram.write()` has not been called
            before `Diagram.compile()`.
        """

        try:
            subprocess.check_call(["latexmk", "--lualatex", f"{self.filename}.tex", "--silent"])
            subprocess.check_call(["latexmk", "-c", f"{self.filename}.tex"])
            if clear_source:
                subprocess.check_call(["rm", f"{self.filename}.tex"])
        except subprocess.CalledProcessError as err:
            if err.returncode == 11:
                raise FileNotFoundError(
                    f"File {self.filename} not found. "
                    "Check if call to Class method write() is missing."
                ) from err

        output_files: list[str] = []

        if not isinstance(output, list):
            output = [output]

        if "svg" in output:
            output_files.append(self._to_svg())
        if "png" in output:
            output_files.append(self._to_png())
        if "pdf" not in output:
            subprocess.check_call(["rm", f"{self.filename}.pdf"])
        else:
            output_files.append(f"{self.filename}.pdf")

        return output_files



"""Simple RBD example.

Code comments:
1.  The arrow style and font can be set by changing the value of
    `config.ARROW_STYLE` and `config.SERIF_FONT`
2.  Here, a `Group` block is made by simply multiplying a `Block` instance by an integer
3.  To group different blocks vertically and set a title and color, use the `Group` class
4.  Adding `Block` instances creates a `Series` instance
5.  To group different blocks horizontally and set a title and color, use the `Series` class
6.  Custom color `myblue` used in first block is defined here

"""

# pylint: disable=C0103
from os import chdir, path

from pyrbd import Block, Diagram, Group, Series, config

config.ARROW_STYLE = "-latex"  # (1)
config.SERIF_FONT = True

chdir(path.dirname(__file__))

# Define all the blocks in the diagram
start_block = Block("Start", "myblue", parent=None)
parallel = 2 * Block("Parallel blocks", "gray", parent=start_block)  # (2)
block_1 = Block(r"Block 1", "yellow!50")
block_2 = Block(r"Block 2", "yellow!50")
block_3 = Block(r"Block 3", "yellow!50")
block_4 = Block(r"Block 4", "yellow!50")
group = Group(  # (3)
    [block_1 + block_2, block_3 + block_4],  # (4)
    parent=parallel,
    text="Group",
    color="yellow",
)
block_a = Block(r"Block A", "orange!50")
block_b = Block(r"Block B", "orange!50")
series = Series([block_a, block_b], "Series", "orange", parent=group)  # (5)
end_block = Block("End", "green!50", parent=series)

# Add blocks to Diagram class instance and compile diagram
diag = Diagram(
    "example_RBD",
    blocks=[start_block, parallel, group, series, end_block],
    hazard="Hazard",
    colors={"myblue": "8888ff"},  # (6)
)
diag.write()
diag.compile(["svg", "png"])



"""Example with layered Series and Group instances."""

# pylint: disable=C0103
from copy import deepcopy
from os import chdir, path

from pyrbd import Block, Diagram, Group, Series

chdir(path.dirname(__file__))

# Define the blocks comprising the diagram
start_block = Block("Start", "blue!30")
block = Block("Block", "gray!20")
group_1 = Group(
    [b * 2 for b in 3 * [block]],
    text="Group with Series",
    color="orange",
    parent=start_block,
)
series_1 = Series(
    [deepcopy(block), 2 * block],
    text="Series with Group",
    color="red",
    parent=group_1,
)
series_2 = Series(
    [2 * block, deepcopy(block), 3 * (block * 2)],
    text="Series with mixed Groups",
    color="RoyalBlue",
    parent=series_1,
)
group_2 = Group(
    [
        block * 3,
        deepcopy(block),
        block * 4,
    ],
    text="Group with mixed Series",
    color="red",
    parent=series_2,
)
series_in_series = Series(
    [
        Series([deepcopy(block), deepcopy(block)], color="yellow", text="Child 1"),
        Series([deepcopy(block), deepcopy(block)], color="orange", text="Child 2"),
        deepcopy(block),
    ],
    parent=group_2,
    text="Parent series",
    color="red",
)
end_block = 2 * Block("End", "green!50", parent=series_in_series)

# Define and compile the diagram
diagram = Diagram(
    "layered_RBD",
    blocks=[
        start_block,
        group_1,
        series_1,
        series_2,
        group_2,
        series_in_series,
        end_block,
    ],
)
diagram.write()
diagram.compile(["pdf", "svg"])



"""Tests for classes in block.py"""

from copy import deepcopy

import pytest

from pyrbd import Block, Group, Series


def test_block() -> None:
    """Test `Block` __init__ and properties."""

    block = Block("Block", "blue")
    tikz_node = block.get_node()
    assert "right=" not in tikz_node
    assert "\\draw" not in tikz_node

    child = Block("Child", "green", shift=(0.5, 2), parent=block)
    assert child.id == str(int(block.id) + 1)
    tikz_node = child.get_node(connector_position=0.5)
    assert f"[right=1.0cm of {block.id}, yshift=2cm]" in tikz_node
    assert (
        f"[arrowcolor, thick, rectangle connector=0.5cm] ({block.id}.east) to ({child.id}.west);"
    ) in tikz_node

    node = child.get_node()
    assert "green" in node
    assert "Child" in node


def test_series() -> None:
    """Test `Series` __init__ and properties."""

    block_1 = Block("Block 1", "blue")
    block_2 = Block("Block 2", "green")
    series = Series([block_1, block_2])

    assert series.parent is None

    assert block_1.id == f"{series.id}+0"
    assert block_2.id == f"{series.id}+1"
    assert block_1.parent is None
    assert block_2.parent is block_1

    tikz_node = series.get_node()

    assert "background" not in tikz_node
    assert "% label text" not in tikz_node

    series_node = Series([block_1, block_2], "Series label", "gray", parent=series)

    assert series_node.id == str(int(series.id) + 1)
    tikz_node = series_node.get_node()
    assert "gray" in tikz_node
    assert "gray!50" in tikz_node and "Series label" in tikz_node

    for color in ["blue", "green", "gray"]:
        assert color in series_node.get_node()


def test_group() -> None:
    """Test `Group` __init__ and properties."""

    block_1 = Block("Block 1", "blue")
    block_2 = Block("Block 2", "green")
    group = Group([block_1, block_2])

    assert group.parent is None

    tikz_node = group.get_node()

    assert "background" not in tikz_node
    assert "% label text" not in tikz_node

    assert len(group.shift) == 2
    assert group.shifts == [0, -group.shift_scale]

    assert block_1.id == f"{group.id}-0"
    assert block_2.id == f"{group.id}-1"
    assert block_1.parent is group
    assert block_2.parent is group

    group_node = Group([block_1, block_2], parent=group)
    assert group_node.id == str(int(group.id) + 1)

    for color in ["blue", "green"]:
        assert color in group_node.get_node()

    group_w_background = Group([block_1, block_2], parent=group, text="Group label", color="black")
    tikz_node = group_w_background.get_node()
    assert "black" in tikz_node
    assert "Group label" in tikz_node


def test_add() -> None:
    """Tests for __add__ for `Block` class."""

    block = Block("block", "white")

    assert isinstance(series := block + deepcopy(block), Series)
    assert len(series.blocks) == 2

    for variable in [2, None, False, 3.0]:
        with pytest.raises(TypeError):
            print(block + variable)  # type: ignore
        with pytest.raises(TypeError):
            print(variable + block)  # type: ignore


def test_mul_rmul() -> None:
    """Tests for __mul__ for `Block` class."""

    block = Block("block", "white")

    assert isinstance(group := 3 * block, Group)
    assert len(group.blocks) == 3

    assert isinstance(series := block * 2, Series)
    assert len(series.blocks) == 2

    assert isinstance(series * 3, Series)
    assert isinstance(3 * series, Group)
    assert isinstance(group * 3, Series)
    assert isinstance(3 * group, Group)

    for value in [-1, 0, 2.4, False]:
        with pytest.raises(ValueError):
            print(value * block)  # type: ignore
        with pytest.raises(ValueError):
            print(block * value)  # type: ignore

    rblock = 1 * block
    assert rblock is block
    assert isinstance(rblock, Block)
    lblock = block * 1
    assert lblock is block
    assert isinstance(lblock, Block)


def test_get_blocks() -> None:
    """Test get_blocks generator function."""

    block = Block("self", "black")
    assert block.get_blocks().__next__() is block

    series = block * 3
    assert list(series.get_blocks()) == series.blocks

    group = 5 * block
    assert list(group.get_blocks()) == group.blocks



"""Package for creating simple reliability block diagrams using LaTeX TikZ."""

from . import config
from .block import Block, Group, Series
from .diagram import Diagram

__all__ = [
    "config",
    "Block",
    "Series",
    "Group",
    "Diagram",
]



"""Simple RBD example.

Annotations:
1.  The first block does not need a parent
2.  To get identical parallel blocks, multiply a `Block` instance by an `int`
3.  For the other blocks we must specify the previous block in the diagram as the parent block
4.  We can specify multiple output formats by providing a list of formats.
    Defaults to PDF if unspecified.
"""

# pylint: disable=C0103
from os import chdir, path

from pyrbd import Block, Diagram

chdir(path.dirname(__file__))

# Define the blocks comprising the diagram
start_block = Block("Start", "blue!30")  # (1)
parallel = 2 * Block("Parallel blocks", "gray", parent=start_block)  # (2)
end_block = Block("End", "green!50", parent=parallel)  # (3)

# Define and compile the diagram
diagram = Diagram(
    "simple_RBD",
    blocks=[start_block, parallel, end_block],
)
diagram.write()
diagram.compile(["pdf", "svg"])  # (4)



"""Module containing Block, Series and Group class definitions."""

from __future__ import annotations

import itertools
from collections import namedtuple
from collections.abc import Generator
from copy import deepcopy
from typing import Literal, overload

from .templates import JINJA_ENV

Padding = namedtuple("Padding", ["n", "e", "s", "w"])


class Block:
    """Block entering a reliability block diagram.

    Parameters
    ----------
    text : str
        block text string
    color : str
        block color
    parent : Block | None, default=None
        parent `Block` instance
    shift : tuple[float, float], default=(0.0, 0.0)
        additional position shift `(x, y)` relative to `parent` `Block` instance

    Attributes
    ----------
    node_options : str
        TikZ node formatting options
    arrow_options : str, default="arrowcolor, thick"
        TikZ arrow formatting options
    arrow_length : float, default=0.5
        default arrow length between nodes (in cm)
    """

    _template = "block.tex.jinja"
    node_options: str = ", ".join(
        [
            "anchor=west",
            "align=center",
            "fill={fill_color}",
            "draw=black!70!gray",
            "minimum height=1cm",
            "rounded corners=0.3mm",
            "inner sep=4pt",
            "outer sep=0pt",
        ]
    )

    arrow_options: str = "arrowcolor, thick"
    arrow_length: float = 0.5
    # draw_arrow: bool = True

    _block_count = itertools.count(start=1)

    def __init__(
        self,
        text: str,
        color: str,
        parent: Block | None = None,
        shift: tuple[float, float] = (0.0, 0.0),
    ) -> None:
        self.text = text
        self.color = color
        self.parent = parent
        self.shift = shift
        self.last = self
        self.id = str(next(self._block_count))

    def get_node(self, connector_position: float | None = None) -> str:
        """Get TikZ node string.

        Parameters
        ----------
        connector_position : float | None, default=None
            distance in cm to right angle bend in connector. Defaults to `0.5*arrow_length`.

        Returns
        -------
        str
            TikZ string for rendering block
        """

        if connector_position is None:
            connector_position = self.arrow_length / 2

        environment = JINJA_ENV
        template = environment.get_template(self._template)
        context = {
            "type": "Block",
            "block": self,
            "node_options": self.node_options.format(fill_color=self.color),
            "connector_position": connector_position,
        }

        return template.render(context)

    def get_blocks(self) -> Generator[Block, None, None]:
        """Yield child `Block` istances."""

        yield self

    def __add__(self, block: Block) -> Series:
        """Add two `Block` instances to make a `Series` instance.

        Parameters
        ----------
        block : Block
            another `Block` instance

        Returns
        -------
        Series
            `Series` instance with `blocks = [self, block]`

        Raises
        ------
        TypeError
            If `block` is not an instance of `Block`
        """

        if not isinstance(block, Block):
            raise TypeError(f"Cannot add object of type {type(block)=} to Block instance.")

        return Series([self, block], parent=self.parent)

    @overload
    def __rmul__(self, value: Literal[1]) -> Block: ...

    @overload
    def __rmul__(self, value: int) -> Group: ...

    def __rmul__(self, value: int) -> Group | Block:
        """Right multiply `Block` instance by `value` to make `Group` with repeated blocks.

        Parameters
        ----------
        value : int
            multiplicative factor, a positive integer

        Returns
        -------
        Group | Block
            `Group` instance with `value` copies of block if `value > 1`, `self` otherwise

        Raises
        ------
        ValueError
            If `value` is not a positive integer
        """

        if not isinstance(value, int) or value <= 0:
            raise ValueError("Multiplicative factor `value` must be a positive integer")

        if value == 1:
            return self

        blocks: list[Block] = [deepcopy(self) for _ in range(value)]

        return Group(blocks, parent=self.parent)

    @overload
    def __mul__(self, value: Literal[1]) -> Block: ...

    @overload
    def __mul__(self, value: int) -> Series: ...

    def __mul__(self, value: int) -> Series | Block:
        """Multiply `Block` instance by `value` to make `Series` with repeated blocks.

        Parameters
        ----------
        value : int
            multiplicative factor, a positive integer

        Returns
        -------
        Series | Block
            `Series` instance with `value` copies of block if `value > 1`, self otherwise

        Raises
        ------
        ValueError
            If `value` is not a positive integer
        """

        if not isinstance(value, int) or value <= 0:
            raise ValueError("Multiplicative factor `value` must be a positive integer")

        if value == 1:
            return self

        blocks: list[Block] = [deepcopy(self) for _ in range(value)]

        return Series(blocks, parent=self.parent)


class Series(Block):
    """Series configuration of `Block` instances for horisontal grouping.

    Parameters
    ----------
    blocks : list[Block]
        list of `Block` instances
    text: str, default=""
        series label text
    color: str, default="white"
        series color, defaults to white
    parent : Block | None, default=None
        parent `Block` instance

    Attributes
    ----------
    node_options : str
        TikZ node options
    internal_arrow_length : float, default=0.3
        distance between blocks in series
    pad : Padding, default=Padding(1, 1, 1, 2.5)
        `namedtuple` `(north, east, south, west)` defining padding (in mmm) between
        blocks and series frame
    label_height : float, default=5.0
        height of series label (in mm)

    """

    _template = "series.tex.jinja"
    node_options: str = ", ".join(
        [
            "anchor=west",
            "align=center",
            "inner sep=0pt",
            "outer sep=0pt",
        ]
    )
    internal_arrow_length: float = 0.3
    pad: Padding = Padding(1, 1, 1, 2.5)
    label_height: float = 5.0

    def __init__(
        self,
        blocks: list[Block],
        text: str = "",
        color: str = "white",
        parent: Block | None = None,
    ) -> None:
        Block.__init__(self, text, color, parent)

        self.blocks = blocks
        self.last = self.blocks[-1]

        self.blocks[0].id = f"{self.id}+0"
        for i, (block, new_parent) in enumerate(
            zip(self.blocks[1::], self.blocks[0:-1], strict=True), start=1
        ):
            block.parent = new_parent
            block.id = f"{self.id}+{i}"
            if not isinstance(block, Series):
                block.arrow_length = self.internal_arrow_length

        if True in [isinstance(block, Series) for block in self.blocks]:
            self.shift = (0, 0.25)

    def get_node(self, connector_position: float | None = None) -> str:
        """Get TikZ node string.

        Parameters
        ----------
        connector_position : float | None, default=None
            distance in cm to right angle bend in connector. Defaults to `0.5 * arrow_length`

        Returns
        -------
        str
            TikZ string for rendering series

        """
        if connector_position is None:
            connector_position = self.arrow_length / 2

        block_nodes = [block.get_node(connector_position) for block in self.blocks]

        environment = JINJA_ENV
        template = environment.get_template(self._template)
        context = {
            "type": "Series",
            "block": self,
            "node_options": self.node_options,
            "connector_position": connector_position,
            "pad": self.pad,
            "block_nodes": block_nodes,
            "first": self.blocks[0],
        }

        return template.render(context)

    def get_blocks(self) -> Generator[Block, None, None]:
        yield from [children for block in self.blocks for children in block.get_blocks()]


class Group(Block):
    """Group of `Block` instances for vertical stacking.

    Parameters
    ----------
    blocks : list[Block]
        list of `Block` instances
    text : str, default=""
        group label text
    color : str, default=""
        group color, defaults to white
    parent : Block | None, default=None
        parent `Block` instance

    Attributes
    ----------
    shift_scale : float, default=1.2
        scaling factor for vertical shifts of blocks
    node_options : str
        TikZ node options
    internal_arrow_length : float, default=0.3
        distance between blocks in series
    end_arrow_scaling : float, default=0.75
        scale factor of last arrow in `Group`
    pad : Padding, default=Padding(1, 1, 1, 1)
        `namedtuple` `(north, east, south, west)` defining padding (in mmm) between
        blocks and series frame
    label_height : float, default=5.0
        height of series label (in mm)
    """

    _template = "group.tex.jinja"
    shift_scale: float = 1.2
    node_options: str = ", ".join(
        [
            "anchor=west",
            "align=center",
            "inner sep=0pt",
            "outer sep=0pt",
        ]
    )
    internal_arrow_length: float = 0.3
    end_arrow_scaling: float = 0.75
    pad: Padding = Padding(1, 1, 1, 1)
    label_height: float = 5.0

    def __init__(
        self,
        blocks: list[Block],
        text: str = "",
        color: str = "",
        parent: Block | None = None,
    ) -> None:
        Block.__init__(self, text, color, parent)

        self.blocks = blocks
        for i, (block, shift) in enumerate(zip(self.blocks, self.shifts, strict=True)):
            block.shift = (0, shift)
            block.parent = self
            block.id = f"{self.id}-{i}"
            block.arrow_length = self.internal_arrow_length
            # block.draw_arrow = False

    @property
    def shifts(self) -> list[float]:
        """List of vertical position shifts for each `Block` instance in group.

        Returns
        -------
        list[float]
            list of vertical position shifts for each `Block` instance in group
        """

        n_blocks = len(self.blocks)

        return list(-self.shift_scale * n for n in range(n_blocks))

    @property
    def sorted_blocks(self) -> list[Block]:
        """Get TikZ string for arrow connecting stacked blocks."""

        # scaling = 0.75

        series_blocks = [block for block in self.blocks if isinstance(block, Series)]
        series_blocks.sort(key=lambda block: len(list(block.get_blocks())), reverse=True)

        longest_series_index = self.blocks.index(series_blocks[0]) if len(series_blocks) > 0 else 0
        blocks = deepcopy(self.blocks)
        longest_series = blocks.pop(longest_series_index)

        return [longest_series, *blocks]

    def get_node(self, connector_position: float | None = None) -> str:
        """Get TikZ node string.

        Parameters
        ----------
        connector_position : float | None, default=None
            distance in cm to right angle bend in connector.
            Locked to 0.0 for `Group` class

        Returns
        -------
        str
            TikZ string for rendering group
        """

        connector_position = 0.0

        block_nodes = [block.get_node(connector_position) for block in self.blocks]

        environment = JINJA_ENV
        template = environment.get_template(self._template)
        context = {
            "type": "Group",
            "block": self,
            "node_options": self.node_options,
            "connector_position": connector_position,
            "pad": self.pad,
            "block_nodes": block_nodes,
        }

        return template.render(context)

    def get_blocks(self) -> Generator[Block, None, None]:
        yield from self.blocks



"""Templates subpackage."""

from os.path import abspath, dirname, join

from jinja2 import Environment, FileSystemLoader, StrictUndefined

_TEMPLATE_DIR = abspath(join(dirname(__file__), "templates/"))
JINJA_ENV: Environment = Environment(
    loader=FileSystemLoader(_TEMPLATE_DIR),
    block_start_string=r"\BLOCK{",
    block_end_string=r"}",
    variable_start_string=r"\VAR{",
    variable_end_string=r"}",
    comment_start_string=r"\#{",
    comment_end_string=r"}",
    line_statement_prefix=r"%%",
    line_comment_prefix=r"%#",
    trim_blocks=True,
    autoescape=False,
    keep_trailing_newline=True,
    undefined=StrictUndefined,
)



"""Tests for `Diagram` class."""

from os import chdir

import pytest

from pyrbd import Block, Diagram


@pytest.fixture(name="diagram")
def diagram_fixture() -> Diagram:
    """Diagram pytest fixture."""

    block1 = Block("block1", "white")
    block2 = Block("block2", "white")
    block3 = Block("block3", "white")
    return Diagram("test_diagram_compile", [block1, block2, block3], "Overpressure")


def test_diagram_compile(tmp_path, diagram: Diagram) -> None:
    """Test `Diagram` `write` method."""

    temp_dir = tmp_path / "test_diagram_compile"
    temp_dir.mkdir()
    chdir(temp_dir)

    with pytest.raises(FileNotFoundError):
        diagram.compile()

    diagram.write()
    assert ".pdf" in "\n".join(diagram.compile(clear_source=False))
    assert ".svg" in "\n".join(diagram.compile("svg", clear_source=False))
    assert ".png" in "\n".join(diagram.compile("png", clear_source=False))
    assert ".pdf" not in "\n".join(diagram.compile(["svg", "png"], clear_source=False))
    assert ".pdf" in "\n".join(diagram.compile(["pdf", "svg"]))
